@@ -19,6 +19,11 @@ def define_env(env):
 ```cpp
 target_link_libraries(myapp QCoro::{module})
 ```
+""")
+        out += row("QMake", f"""
+```cpp
+QT += QCoro{module}
+```
 """)
         if inherits:
             out += row("Inherits", inheritsLink(inherits))
