@@ -5,6 +5,7 @@
 #include "testobject.h"
 #include "qcorotask.h"
 #include "qcorotimer.h"
+#include "qcorosignal.h"
 
 #include <QTest>
 #include <QObject>
@@ -57,6 +58,66 @@ public:
     QString string;
 };
 
+struct TestAwaitableBase {
+    std::chrono::milliseconds delay() const { return mDelay; }
+private:
+    std::chrono::milliseconds mDelay = 100ms;
+};
+
+template<typename T>
+struct TestAwaitable : TestAwaitableBase {
+public:
+    TestAwaitable(T val)
+        : mResult(val)
+    {}
+
+    bool await_ready() const { return false; }
+    void await_suspend(std::coroutine_handle<> handle) {
+        QTimer::singleShot(100ms, [handle = std::move(handle)]() {
+            handle.resume();
+        });
+    }
+    T await_resume() {
+        return mResult;
+    }
+
+private:
+    T mResult;
+};
+
+template<>
+struct TestAwaitable<void> : TestAwaitableBase {
+public:
+    bool await_ready() const { return false; }
+    void await_suspend(std::coroutine_handle<> handle) {
+        QTimer::singleShot(100ms, [handle = std::move(handle)]() {
+            handle.resume();
+        });
+    }
+    void await_resume() {}
+};
+
+template<typename T>
+struct TestAwaitableWithCoAwait {
+    TestAwaitableWithCoAwait(T val)
+        : mResult(val)
+    {}
+
+    TestAwaitable<T> operator co_await() {
+        return TestAwaitable(mResult);
+    }
+
+private:
+    T mResult;
+};
+
+template<>
+struct TestAwaitableWithCoAwait<void> {
+    TestAwaitable<void> operator co_await() {
+        return TestAwaitable<void>();
+    }
+};
+
 } // namespace
 
 class QCoroTaskTest : public QCoro::TestObject<QCoroTaskTest>
@@ -334,6 +395,73 @@ private:
         QCOMPARE(result, QStringLiteral("42"));
     }
 
+    QCoro::Task<> testMultipleAwaiters_coro(QCoro::TestContext) {
+        auto task = timer(100ms);
+
+        bool called = false;
+        // Internally co_awaits task
+        task.then([&called]() {
+            called = true;
+        });
+
+        co_await task;
+
+        QCORO_VERIFY(called);
+    }
+
+    QCoro::Task<> testMultipleAwaitersSync_coro(QCoro::TestContext ctx) {
+        ctx.setShouldNotSuspend();
+
+        auto task = []() -> QCoro::Task<> { co_return; }();
+
+        bool called = false;
+        task.then([&called]() {
+            called = true;
+        });
+
+        co_await task;
+
+        QCORO_VERIFY(called);
+    }
+
+    Q_SIGNAL void callbackCalled();
+
+    template <typename QObjectDerived, typename Signal>
+    QCoro::Task<> verifySignalEmitted(QObjectDerived *context, Signal &&signal) {
+        bool called = false;
+        co_await qCoro(context, std::move(signal)).then([&]() {
+            called = true;
+        });
+        QCORO_VERIFY(called);
+    }
+
+    QCoro::Task<> testTaskConnect_coro(QCoro::TestContext) {
+        // Test that free functions can be passed as callback
+        QCoro::connect(timer(), this, [this]() {
+            Q_EMIT callbackCalled();
+        });
+        co_await verifySignalEmitted(this, &QCoroTaskTest::callbackCalled);
+
+        // Check that member functions can be passed as callback
+        QCoro::connect(timer(), this, &QCoroTaskTest::callbackCalled);
+        co_await verifySignalEmitted(this, &QCoroTaskTest::callbackCalled);
+
+        // Test that the code still compiles if the value of the coroutine is not used by the function.
+        auto nonVoidCoroutine = []() -> QCoro::Task<QString> {
+            co_await timer();
+            co_return QStringLiteral("Hello World!");
+        };
+        QCoro::connect(nonVoidCoroutine(), this, [this]() {
+            Q_EMIT callbackCalled();
+        });
+        co_await verifySignalEmitted(this, &QCoroTaskTest::callbackCalled);
+
+        QCoro::connect(nonVoidCoroutine(), this, [this](QString) {
+            Q_EMIT callbackCalled();
+        });
+        co_await verifySignalEmitted(this, &QCoroTaskTest::callbackCalled);
+    }
+
 private Q_SLOTS:
     addTest(SimpleCoroutine)
     addTest(CoroutineValue)
@@ -354,7 +482,10 @@ private Q_SLOTS:
     addTest(ThenExceptionPropagation)
     addTest(ThenError)
     addTest(ThenErrorWithValue)
+    addTest(TaskConnect)
     addThenTest(ImplicitArgumentConversion)
+    addTest(MultipleAwaiters)
+    addTest(MultipleAwaitersSync)
 
     // See https://github.com/danvratil/qcoro/issues/24
     void testEarlyReturn()
@@ -418,6 +549,53 @@ private Q_SLOTS:
         QCOMPARE(result, 42);
     }
 
+    void testWaitForAwaitable() {
+        TestAwaitable<int> awaitable(42);
+        QElapsedTimer timer;
+        timer.start();
+
+        static_assert(std::is_same_v<decltype(QCoro::waitFor(awaitable)), int>);
+        const int result = QCoro::waitFor(awaitable);
+        QCOMPARE(result, 42);
+        QVERIFY(timer.elapsed() >= static_cast<float>(awaitable.delay().count()) * 0.9);
+
+    }
+
+    void testWaitForVoidAwaitable() {
+        TestAwaitable<void> awaitable;
+        QElapsedTimer timer;
+        timer.start();
+
+        static_assert(std::is_void_v<decltype(QCoro::waitFor(awaitable))>);
+        QCoro::waitFor(awaitable);
+
+        QVERIFY(timer.elapsed() >= static_cast<float>(awaitable.delay().count()) * 0.9);
+    }
+
+    void testWaitForAwaitableWithOperatorCoAwait() {
+        TestAwaitableWithCoAwait<int> awaitable(42);
+        QCoro::waitFor(awaitable);
+        QElapsedTimer timer;
+        timer.start();
+
+        static_assert(std::is_same_v<decltype(QCoro::waitFor(awaitable)), int>);
+        const int result = QCoro::waitFor(awaitable);
+        QCOMPARE(result, 42);
+        QVERIFY(timer.elapsed() >= (90ms).count());
+    }
+
+    void testWaitForVoidAwaitableWithOperatorCoAwait() {
+        TestAwaitableWithCoAwait<void> awaitable;
+        QElapsedTimer timer;
+        timer.start();
+
+        static_assert(std::is_void_v<decltype(QCoro::waitFor(awaitable))>);
+        QCoro::waitFor(awaitable);
+
+        QVERIFY(timer.elapsed() >= (90ms).count());
+    }
+
+
     void testIgnoredVoidTaskResult() {
         QEventLoop el;
         ignoreCoroutineResult(el, [&el]() -> QCoro::Task<> {
@@ -536,6 +714,27 @@ private Q_SLOTS:
 
         el.exec();
     }
+
+    void testTaskConnectContext_coro() {
+        auto task = timer(200ms);
+        static_assert(std::is_same_v<decltype(task), QCoro::Task<>>);
+
+        bool called = false;
+        auto context = new QObject();
+
+        QCoro::connect(task, context, [&]() {
+            called = true;
+        });
+
+        // Delete context, callback should not be called
+        delete context;
+
+        QCoro::waitFor(task);
+
+        QVERIFY(!called);
+    }
+
+
 };
 
 QTEST_GUILESS_MAIN(QCoroTaskTest)
